# Copyright 2015 Cisco Systems, Inc.  All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import hashlib
import json
import os
import sys
import traceback
kb_main_path = os.path.split(os.path.abspath(__file__))[0] + "/../../../kloudbuster"
sys.path.append(kb_main_path)

from credentials import Credentials
from kb_session import KBSession
from kb_session import KBSessionManager
import log as logging

from configure import Configuration
from kb_config import KBConfig
from pecan import expose
from pecan import response

class ConfigController(object):

    @expose(generic=True)
    def default_config(self):
        kb_config = KBConfig()
        pdict = eval(str(kb_config.config_scale))
        # Normally we don't allow the clients to change below configs
        pdict['client'].pop('tp_tool')
        pdict['client'].pop('http_tool')
        return json.dumps(pdict)

    @expose(generic=True)
    def running_config(self, *args):
        if len(args):
            session_id = args[0]
        else:
            response.status = 400
            response.text = u"Please specify the session_id."
            return response.text
        if KBSessionManager.has(session_id):
            kb_config_obj = KBSessionManager.get(session_id).kb_config
            config_scale = kb_config_obj.config_scale
            config_scale['server'] = kb_config_obj.server_cfg
            config_scale['client'] = kb_config_obj.client_cfg
            config_scale = eval(str(config_scale))
            return json.dumps(config_scale)
        else:
            response.status = 404
            response.text = u"Session ID is not found or invalid."
            return response.text

    @running_config.when(method='POST')
    def running_config_POST(self, arg):
        try:
            # Expectation:
            # {
            #  'credentials': {'tested-rc': '<STRING>', 'tested-passwd': '<STRING>',
            #                  'testing-rc': '<STRING>', 'testing-passwd': '<STRING>'},
            #  'kb_cfg': {<USER_OVERRIDED_CONFIGS>},
            #  'topo_cfg': {<TOPOLOGY_CONFIGS>}
            #  'tenants_cfg': {<TENANT_AND_USER_LISTS_FOR_REUSING>}
            # }
            user_config = json.loads(arg)

            # Parsing credentials from application input
            cred_config = user_config['credentials']
            cred_tested = Credentials(openrc_contents=cred_config['tested-rc'],
                                      pwd=cred_config['tested-passwd'])
            if ('testing-rc' in cred_config and
               cred_config['testing-rc'] != cred_config['tested-rc']):
                cred_testing = Credentials(openrc_contents=cred_config['testing-rc'],
                                           pwd=cred_config['testing-passwd'])
            else:
                # Use the same openrc file for both cases
                cred_testing = cred_tested

            session_id = hashlib.md5(str(cred_config)).hexdigest()
            kb_config = KBConfig()
            if KBSessionManager.has(session_id):
                response.status = 403
                response.text = u"Session is already existed."
                return response.text

            # Parsing server and client configs from application input
            # Save the public key into a temporary file
            if 'public_key' in user_config['kb_cfg']:
                pubkey_filename = '/tmp/kb_public_key.pub'
                f = open(pubkey_filename, 'w')
                f.write(user_config['kb_cfg']['public_key_file'])
                f.close()
                kb_config.config_scale['public_key_file'] = pubkey_filename

            if 'prompt_before_run' in user_config['kb_cfg']:
                kb_config.config_scale['prompt_before_run'] = False

            if user_config['kb_cfg']:
                alt_config = Configuration.from_string(user_config['kb_cfg']).configure()
                kb_config.config_scale = kb_config.config_scale.merge(alt_config)

            # Parsing topology configs from application input
            if 'topo_cfg' in user_config:
                topo_cfg = Configuration.from_string(user_config['topo_cfg']).configure()
            else:
                topo_cfg = None

            # Parsing tenants configs from application input
            if 'tenants_list' in user_config:
                tenants_list = Configuration.from_string(user_config['tenants_list']).configure()
            else:
                tenants_list = None
        except Exception:
            response.status = 400
            response.text = u"Error while parsing configurations: \n%s" % (traceback.format_exc)
            return response.text

        logging.setup("kloudbuster", logfile="/tmp/kb_log_%s" % session_id)
        kb_config.init_with_rest_api(cred_tested=cred_tested,
                                     cred_testing=cred_testing,
                                     topo_cfg=topo_cfg,
                                     tenants_list=tenants_list)

        kb_session = KBSession()
        kb_session.kb_config = kb_config
        KBSessionManager.add(session_id, kb_session)

        return str(session_id)

    @running_config.when(method='PUT')
    def running_config_PUT(self, *args):
        # @TODO(Not completed! ENOTSUP)
        if len(args):
            session_id = args[0]
        else:
            response.status = 400
            response.text = u"Please specify the session_id."
            return response.text
        if KBSessionManager.has(session_id):
            # kb_session = KBSessionManager.get(session_id)
            #
            #
            #
            return "OK!"
        else:
            response.status = 404
            response.text = u"Session ID is not found or invalid."
            return response.text

    @running_config.when(method='DELETE')
    def running_config_DELETE(self, *args):
        if len(args):
            session_id = args[0]
        else:
            response.status = 400
            response.text = u"Please specify the session_id."
            return response.text
        if KBSessionManager.has(session_id):
            kb_session = KBSessionManager.get(session_id)
            if kb_session.kloudbuster:
                kb_session.kloudbuster.dispose()
            KBSessionManager.delete(session_id)
            return "OK!"
        else:
            response.status = 404
            response.text = u"Session ID is not found or invalid."
            return response.text



# Copyright 2015 Cisco Systems, Inc.  All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

import subprocess
import sys
import threading
import time

import redis

# Define the version of the KloudBuster agent and VM image
#
# When VM is up running, the agent will send the READY message to the
# KloudBuster main program, along with its version. The main program
# will check the version to see whether the image meets the minimum
# requirements to run, and stopped with an error if not.
#
# This version must be incremented if the interface changes or if new features
# are added to the agent VM
__version__ = '3'

def get_image_name():
    '''Return the versioned VM image name that corresponds to this
    agent code. This string must match the way DIB names the kloudbuster image.
    Return:
        the versioned image name without the extension ('.qcow2' is implicit)
    '''
    return 'kloudbuster_v' + __version__

def get_image_version():
    return __version__

class KB_Instance(object):

    # Check whether the HTTP Service is up running
    @staticmethod
    def check_http_service(target_url):
        cmd = 'while true; do\n'
        cmd += 'curl --head %s --connect-timeout 2 --silent\n' % (target_url)
        cmd += 'if [ $? -eq 0 ]; then break; fi\n'
        cmd += 'done'
        return cmd

    # Add static route
    @staticmethod
    def add_static_route(network, next_hop_ip, if_name=None):
        debug_msg = "Adding static route %s with next hop %s" % (network, next_hop_ip)
        cmd = "sudo ip route add %s via %s" % (network, next_hop_ip)
        if if_name:
            debug_msg += " and %s" % if_name
            cmd += " dev %s" % if_name
        # TODO(Logging on Agent)
        print debug_msg
        return cmd

    # Get static route
    @staticmethod
    def get_static_route(network, next_hop_ip=None, if_name=None):
        cmd = "ip route show %s" % network
        if next_hop_ip:
            cmd += " via %s" % next_hop_ip
        if if_name:
            cmd += " dev %s" % if_name
        return cmd

    # Delete static route
    @staticmethod
    def delete_static_route(network, next_hop_ip=None, if_name=None):
        debug_msg = "Deleting static route %s" % network
        cmd = "sudo ip route del %s" % network
        if next_hop_ip:
            debug_msg = " with next hop %s" % next_hop_ip
            cmd += " via %s" % next_hop_ip
        if if_name:
            if next_hop_ip:
                debug_msg = " and %s" % if_name
            else:
                debug_msg = "with next hop %s" % if_name
            cmd += " dev %s" % if_name
        # TODO(Logging on Agent)
        print debug_msg
        return cmd

    # Run the HTTP benchmarking tool
    @staticmethod
    def run_http_test(dest_path, target_url, threads, connections,
                      rate_limit, duration, timeout, connection_type,
                      report_interval):
        if not rate_limit:
            rate_limit = 65535

        cmd = '%s -t%d -c%d -R%d -d%ds -p%ds --timeout %ds -D2 -e %s' % \
              (dest_path, threads, connections, rate_limit, duration,
               report_interval, timeout, target_url)
        return cmd


class KB_VM_Agent(object):

    def __init__(self, user_data):
        host = user_data['redis_server']
        port = user_data['redis_server_port']
        self.user_data = user_data
        self.redis_obj = redis.StrictRedis(host=host, port=port)
        self.pubsub = self.redis_obj.pubsub(ignore_subscribe_messages=True)
        self.hello_thread = None
        self.stop_hello = threading.Event()
        self.vm_name = user_data['vm_name']
        self.orches_chan_name = "kloudbuster_orches"
        self.report_chan_name = "kloudbuster_report"
        self.last_cmd = None

    def setup_channels(self):
        # Check for connections to redis server
        while (True):
            try:
                self.redis_obj.get("test")
            except (redis.exceptions.ConnectionError):
                time.sleep(1)
                continue
            break

        # Subscribe to orchestration channel
        self.pubsub.subscribe(self.orches_chan_name)

    def report(self, cmd, client_type, data):
        message = {'cmd': cmd, 'sender-id': self.vm_name,
                   'client-type': client_type, 'data': data}
        self.redis_obj.publish(self.report_chan_name, message)

    def send_hello(self):
        # Sending "hello" message to master node every 2 seconds
        while not self.stop_hello.is_set():
            self.report('READY', None, __version__)
            time.sleep(2)

    def exec_command(self, cmd):
        # Execute the command, and returns the outputs
        cmds = ['bash', '-c']
        cmds.append(cmd)
        p = subprocess.Popen(cmds, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        (stdout, stderr) = p.communicate()

        return (p.returncode, stdout, stderr)

    def exec_command_report(self, cmd):
        # Execute the command, reporting periodically, and returns the outputs
        cmd_res_dict = None
        cmds = ['bash', '-c']
        cmds.append(cmd)
        p_output = ''
        p = subprocess.Popen(cmds, stdout=subprocess.PIPE, stderr=subprocess.PIPE)

        lines_iterator = iter(p.stdout.readline, b"")
        for line in lines_iterator:
            # One exception, if this is the very last report, we will send it
            # through "DONE" command, not "REPORT". So what's happening here
            # is to determine whether this is the last report.
            if cmd_res_dict:
                self.report('REPORT', 'http', cmd_res_dict)
                cmd_res_dict = None
                p_output = line
            else:
                p_output += line
                if line.strip() == "}":
                    cmd_res_dict = dict(zip(("status", "stdout", "stderr"), (0, p_output, '')))
                    continue

        stderr = p.communicate()[1]
        return (p.returncode, p_output, stderr)

    def process_cmd(self, message):
        if message['cmd'] == 'ACK':
            # When 'ACK' is received, means the master node
            # acknowledged the current VM. So stopped sending more
            # "hello" packet to the master node.
            # Unfortunately, there is no thread.stop() in Python 2.x
            self.stop_hello.set()
        elif message['cmd'] == 'EXEC':
            self.last_cmd = ""
            try:
                cmd_res_tuple = eval('self.exec_' + message['data']['cmd'] + '()')
                cmd_res_dict = dict(zip(("status", "stdout", "stderr"), cmd_res_tuple))
            except Exception as exc:
                cmd_res_dict = {
                    "status": 1,
                    "stdout": self.last_cmd,
                    "stderr": str(exc)
                }
            self.report('DONE', message['client-type'], cmd_res_dict)
        elif message['cmd'] == 'ABORT':
            # TODO(Add support to abort a session)
            pass
        else:
            # Unexpected
            # TODO(Logging on Agent)
            print 'ERROR: Unexpected command received!'
            pass

    def work(self):
        for item in self.pubsub.listen():
            if item['type'] != 'message':
                continue
            # Convert the string representation of dict to real dict obj
            message = eval(item['data'])
            self.process_cmd(message)

    def exec_setup_static_route(self):
        self.last_cmd = KB_Instance.get_static_route(self.user_data['target_subnet_ip'])
        result = self.exec_command(self.last_cmd)
        if (self.user_data['target_subnet_ip'] not in result[1]):
            self.last_cmd = KB_Instance.add_static_route(
                self.user_data['target_subnet_ip'],
                self.user_data['target_shared_interface_ip'])
            return self.exec_command(self.last_cmd)
        else:
            return (0, '', '')

    def exec_check_http_service(self):
        self.last_cmd = KB_Instance.check_http_service(self.user_data['target_url'])
        return self.exec_command(self.last_cmd)

    def exec_run_http_test(self):
        self.last_cmd = KB_Instance.run_http_test(
            dest_path=self.user_data['http_tool']['dest_path'],
            target_url=self.user_data['target_url'],
            **self.user_data['http_tool_configs'])
        return self.exec_command_report(self.last_cmd)

def exec_command(cmd):
    p = subprocess.Popen(cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    (stdout, stderr) = p.communicate()

    return p.returncode

def start_redis_server():
    cmd = ['sudo', 'service', 'redis-server', 'start']
    return exec_command(cmd)

def start_nuttcp_server():
    cmd = ['/usr/bin/nuttcp', '-P5002', '-S', '--single-threaded']
    return exec_command(cmd)

def start_nginx_server():
    cmd = ['sudo', 'service', 'nginx', 'start']
    return exec_command(cmd)

if __name__ == "__main__":
    try:
        f = open('user-data', 'r')
        user_data = eval(f.read())
    except Exception as e:
        # TODO(Logging on Agent)
        print e.message
        sys.exit(1)

    if 'role' not in user_data:
        sys.exit(1)

    if user_data['role'] == 'KB-PROXY':
        sys.exit(start_redis_server())
    if user_data['role'] == 'Server':
        rc1 = start_nuttcp_server()
        rc2 = start_nginx_server()
        sys.exit(rc1 or rc2)
    elif user_data['role'] == 'Client':
        agent = KB_VM_Agent(user_data)
        agent.setup_channels()
        agent.hello_thread = threading.Thread(target=agent.send_hello)
        agent.hello_thread.daemon = True
        agent.hello_thread.start()
        agent.work()
    else:
        sys.exit(1)



# Copyright 2015 Cisco Systems, Inc.  All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

import abc

import log as logging

LOG = logging.getLogger(__name__)


# A base class for all tools that can be associated to an instance
class PerfTool(object):
    __metaclass__ = abc.ABCMeta

    def __init__(self, instance, tool_name):
        self.instance = instance
        self.name = tool_name
        self.pid = None

    # Terminate pid if started
    def dispose(self):
        if self.pid:
            # Terminate the server
            LOG.kbdebug("[%s] Terminating %s" % (self.instance.vm_name,
                                                 self.name))
            self.instance.ssh.kill_proc(self.pid)
            self.pid = None

    def parse_error(self, msg):
        return {'error': msg, 'tool': self.name}

    def parse_results(self, protocol=None, throughput=None, lossrate=None, retrans=None,
                      rtt_ms=None, reverse_dir=False, msg_size=None, cpu_load=None,
                      http_total_req=None, http_rps=None, http_tp_kbytes=None,
                      http_sock_err=None, http_sock_timeout=None, http_err=None,
                      latency_stats=None):
        res = {'tool': self.name}
        if throughput is not None:
            res['throughput_kbps'] = throughput
        if protocol is not None:
            res['protocol'] = protocol
        if 'vm_bandwidth' in self.instance.config:
            res['bandwidth_limit_kbps'] = self.instance.config.vm_bandwidth
        if lossrate is not None:
            res['loss_rate'] = lossrate
        if retrans:
            res['retrans'] = retrans
        if rtt_ms:
            res['rtt_ms'] = rtt_ms
        if reverse_dir:
            res['direction'] = 'reverse'
        if msg_size:
            res['pkt_size'] = msg_size
        if cpu_load:
            res['cpu_load'] = cpu_load
        if http_total_req:
            res['http_total_req'] = http_total_req
        if http_rps:
            res['http_rps'] = http_rps
        if http_tp_kbytes:
            res['http_throughput_kbytes'] = http_tp_kbytes
        if http_sock_err:
            res['http_sock_err'] = http_sock_err
        if http_sock_timeout:
            res['http_sock_timeout'] = http_sock_timeout
        if http_err:
            res['http_err'] = http_err
        if latency_stats:
            res['latency_stats'] = latency_stats
        return res

    @abc.abstractmethod
    def cmd_parser_run_client(self, status, stdout, stderr):
        # must be implemented by sub classes
        return None

    @staticmethod
    @abc.abstractmethod
    def consolidate_results(results):
        # must be implemented by sub classes
        return None

    def find_udp_bdw(self, pkt_size, target_ip):
        '''Find highest UDP bandwidth within max loss rate for given packet size
        :return: a dictionary describing the optimal bandwidth (see parse_results())
        '''
        # we use a binary search to converge to the optimal throughput
        # start with 5Gbps - mid-range between 1 and 10Gbps
        # Convergence can be *very* tricky because UDP throughput behavior
        # can vary dramatically between host runs and guest runs.
        # The packet rate limitation is going to dictate the effective
        # send rate, meaning that small packet sizes will yield the worst
        # throughput.
        # The measured throughput can be vastly smaller than the requested
        # throughput even when the loss rate is zero when the sender cannot
        # send fast enough to fill the network, in that case increasing the
        # requested rate will not make it any better
        # Examples:
        # 1. too much difference between requested/measured bw - regardless of loss rate
        #    => retry with bw mid-way between the requested bw and the measured bw
        # /tmp/nuttcp-7.3.2 -T2  -u -l128 -R5000000K -p5001 -P5002 -fparse 192.168.1.2
        # megabytes=36.9785 real_seconds=2.00 rate_Mbps=154.8474 tx_cpu=23 rx_cpu=32
        #         drop=78149 pkt=381077 data_loss=20.50746
        # /tmp/nuttcp-7.3.2 -T2  -u -l128 -R2500001K -p5001 -P5002 -fparse 192.168.1.2
        # megabytes=47.8063 real_seconds=2.00 rate_Mbps=200.2801 tx_cpu=24 rx_cpu=34
        #         drop=0 pkt=391629 data_loss=0.00000
        # 2. measured and requested bw are very close :
        #   if loss_rate is too low
        #     increase bw mid-way between requested and last max bw
        #   if loss rate is too high
        #     decrease bw mid-way between the measured bw and the last min bw
        #   else stop iteration (converged)
        # /tmp/nuttcp-7.3.2 -T2  -u -l8192 -R859376K -p5001 -P5002 -fparse 192.168.1.2
        # megabytes=204.8906 real_seconds=2.00 rate_Mbps=859.2992 tx_cpu=99 rx_cpu=10
        #           drop=0 pkt=26226 data_loss=0.00000

        min_kbps = 1
        max_kbps = 10000000
        kbps = 5000000
        min_loss_rate = self.instance.config.udp_loss_rate_range[0]
        max_loss_rate = self.instance.config.udp_loss_rate_range[1]
        # stop if the remaining range to cover is less than 5%
        while (min_kbps * 100 / max_kbps) < 95:
            res_list = self.run_client_dir(target_ip, 0, bandwidth_kbps=kbps,
                                           udp=True, length=pkt_size,
                                           no_cpu_timed=1)
            # always pick the first element in the returned list of dict(s)
            # should normally only have 1 element
            res = res_list[0]
            if 'error' in res:
                return res
            loss_rate = res['loss_rate']
            measured_kbps = res['throughput_kbps']
            LOG.kbdebug(
                "[%s] pkt-size=%d throughput=%d<%d/%d<%d Kbps loss-rate=%d" %
                (self.instance.vm_name, pkt_size, min_kbps, measured_kbps,
                 kbps, max_kbps, loss_rate))
            # expected rate must be at least 80% of the requested rate
            if (measured_kbps * 100 / kbps) < 80:
                # the measured bw is too far away from the requested bw
                # take half the distance or 3x the measured bw whichever is lowest
                kbps = measured_kbps + (kbps - measured_kbps) / 2
                if measured_kbps:
                    kbps = min(kbps, measured_kbps * 3)
                max_kbps = kbps
                continue
            # The measured bw is within striking distance from the requested bw
            # increase bw if loss rate is too small
            if loss_rate < min_loss_rate:
                # undershot
                if measured_kbps > min_kbps:
                    min_kbps = measured_kbps
                else:
                    # to make forward progress we need to increase min_kbps
                    # and try a higher bw since the loss rate is too low
                    min_kbps = int((max_kbps + min_kbps) / 2)

                kbps = int((max_kbps + min_kbps) / 2)
                # LOG.info("   undershot, min=%d kbps=%d max=%d" % (min_kbps,  kbps, max_kbps))
            elif loss_rate > max_loss_rate:
                # overshot
                max_kbps = kbps
                if measured_kbps < kbps:
                    kbps = measured_kbps
                else:
                    kbps = int((max_kbps + min_kbps) / 2)
                # LOG.info("   overshot, min=%d kbps=%d max=%d" % (min_kbps,  kbps, max_kbps))
            else:
                # converged within loss rate bracket
                break
        return res



# Copyright 2015 Cisco Systems, Inc.  All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import json
import os
import sys
import threading
kb_main_path = os.path.split(os.path.abspath(__file__))[0] + "/../../../kloudbuster"
sys.path.append(kb_main_path)

from kb_session import KBSessionManager
from kloudbuster import __version__ as kb_version
from kloudbuster import KloudBuster

from pecan import expose
from pecan import response

class KBController(object):

    def __init__(self):
        self.kb_thread = None

    def kb_thread_handler(self, session_id):
        kb_session = KBSessionManager.get(session_id)
        kb_session.kb_status = 'RUNNING'
        kb_config = kb_session.kb_config
        try:
            kloudbuster = KloudBuster(
                kb_config.cred_tested, kb_config.cred_testing,
                kb_config.server_cfg, kb_config.client_cfg,
                kb_config.topo_cfg, kb_config.tenants_list)
            kb_session.kloudbuster = kloudbuster

            if kloudbuster.check_and_upload_images():
                kloudbuster.run()
            kb_session.kb_status = 'READY'
        except Exception:
            kb_session.kb_status = 'ERROR'

    @expose(generic=True)
    def status(self, *args):
        if len(args):
            session_id = args[0]
        else:
            response.status = 400
            response.text = u"Please specify the session_id."
            return response.text

        if KBSessionManager.has(session_id):
            status = KBSessionManager.get(session_id).kb_status
            return status
        else:
            response.status = 404
            response.text = u"Session ID is not found or invalid."
            return response.text

    @expose(generic=True)
    def log(self, *args, **kwargs):
        if len(args):
            session_id = args[0]
        else:
            response.status = 400
            response.text = u"Please specify the session_id."
            return response.text

        offset = 0
        if 'offset' in kwargs:
            try:
                offset = int(kwargs['offset'])
            except ValueError:
                response.status = 400
                response.text = u"Parameter 'offset' is invalid."
                return response.text

        if KBSessionManager.has(session_id):
            kb_session = KBSessionManager.get(session_id)
            plog = kb_session.kloudbuster.dump_logs(offset=offset)\
                if kb_session.kloudbuster else ""
            return json.dumps(plog)
        else:
            response.status = 404
            response.text = u"Session ID is not found or invalid."
            return response.text

    @expose(generic=True)
    def report(self, *args, **kwargs):
        if len(args):
            session_id = args[0]
        else:
            response.status = 400
            response.text = u"Please specify the session_id."
            return response.text

        final = False
        if 'final' in kwargs:
            final = True if kwargs['final'].lower() == 'true' else False

        preport = None
        if KBSessionManager.has(session_id):
            kb_session = KBSessionManager.get(session_id)
            if kb_session.kloudbuster and kb_session.kloudbuster.kb_runner:
                preport = kb_session.kloudbuster.final_result\
                    if final else kb_session.kloudbuster.kb_runner.report

            return json.dumps(preport)
        else:
            response.status = 404
            response.text = u"Session ID is not found or invalid."
            return response.text

    @expose(generic=True)
    def version(self):
        return kb_version

    @expose(generic=True)
    def run(self, *args):
        response.status = 400
        response.text = u"Please POST to this resource."
        return response.text

    @run.when(method='POST')
    def run_POST(self, *args):
        if len(args):
            session_id = args[0]
        else:
            response.status = 400
            response.text = u"Please specify the session_id."
            return response.text
        if not KBSessionManager.has(session_id):
            response.status = 404
            response.text = u"Session ID is not found or invalid."
            return response.text
        if KBSessionManager.get(session_id).kb_status == 'RUNNING':
            response.status = 403
            response.text = u"An instance of KloudBuster is already running."
            return response.text

        self.kb_thread = threading.Thread(target=self.kb_thread_handler, args=[session_id])
        self.kb_thread.daemon = True
        self.kb_thread.start()

        return "OK!"



# Copyright 2015 Cisco Systems, Inc.  All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import base_compute
import base_network
from cinderclient.v2 import client as cinderclient
import keystoneclient.openstack.common.apiclient.exceptions as keystone_exception
import log as logging
from neutronclient.v2_0 import client as neutronclient
from novaclient.client import Client

LOG = logging.getLogger(__name__)

class KBFlavorCheckException(Exception):
    pass

class KBQuotaCheckException(Exception):
    pass

class User(object):
    """
    User class that stores router list
    Creates and deletes N routers based on num of routers
    """

    def __init__(self, user_name, password, tenant, user_role):
        """
        Store all resources
        1. Keystone client object
        2. Tenant and User information
        3. nova and neutron clients
        4. router list
        """
        self.user_name = user_name
        self.password = password
        self.tenant = tenant
        self.res_logger = tenant.res_logger
        self.router_list = []
        # Store the nova, neutron and cinder client
        self.nova_client = None
        self.neutron_client = None
        self.cinder_client = None
        # Each user is associated to 1 key pair at most
        self.key_pair = None
        self.key_name = None

        # Create the user within the given tenant associate
        # admin role with user. We need admin role for user
        # since we perform VM placement in future
        #
        # If running on top of existing tenants/users, skip
        # the step for admin role association.
        if not self.tenant.reusing_users:
            self.user = self._get_user()
            current_role = self.tenant.kloud.keystone.roles.find(name=user_role)
            self.tenant.kloud.keystone.roles.add_user_role(self.user,
                                                           current_role,
                                                           tenant.tenant_id)
        else:
            # Only admin can retrive the object via Keystone API
            self.user = None
            LOG.info("Using user: " + self.user_name)


    def _create_user(self):
        LOG.info("Creating user: " + self.user_name)
        return self.tenant.kloud.keystone.users.create(name=self.user_name,
                                                       password=self.password,
                                                       email="kloudbuster@localhost",
                                                       tenant_id=self.tenant.tenant_id)

    def _get_user(self):
        '''
        Create a new user or reuse if it already exists (on a different tenant)
        delete the user and create a new one
        '''
        try:
            user = self._create_user()
            return user
        except keystone_exception.Conflict as exc:
            # Most likely the entry already exists (leftover from past failed runs):
            # Conflict: Conflict occurred attempting to store user - Duplicate Entry (HTTP 409)
            if exc.http_status != 409:
                raise exc
            # Try to repair keystone by removing that user
            LOG.warn("User creation failed due to stale user with same name: " +
                     self.user_name)
            user = self.tenant.kloud.keystone.users.find(name=self.user_name)
            LOG.info("Deleting stale user with name: " + self.user_name)
            self.tenant.kloud.keystone.users.delete(user)
            return self._create_user()

        # Should never come here
        raise Exception()

    def delete_resources(self):
        LOG.info("Deleting all user resources for user %s" % self.user_name)

        flag = True
        # Delete key pair
        if self.key_pair:
            self.key_pair.remove_public_key()

        # Delete all user routers
        for router in self.router_list:
            flag = flag & router.delete_router()

        if not self.tenant.reusing_users:
            # Finally delete the user
            self.tenant.kloud.keystone.users.delete(self.user.id)

        return flag

    def update_tenant_quota(self, tenant_quota):
        nova_quota = base_compute.NovaQuota(self.nova_client, self.tenant.tenant_id)
        nova_quota.update_quota(**tenant_quota['nova'])

        cinder_quota = base_compute.CinderQuota(self.cinder_client, self.tenant.tenant_id)
        cinder_quota.update_quota(**tenant_quota['cinder'])

        neutron_quota = base_network.NeutronQuota(self.neutron_client, self.tenant.tenant_id)
        neutron_quota.update_quota(tenant_quota['neutron'])

    def check_resources_quota(self):
        # Flavor check
        flavor_manager = base_compute.Flavor(self.nova_client)
        flavor_to_use = None
        for flavor in flavor_manager.list():
            flavor = vars(flavor)
            if flavor['vcpus'] < 1 or flavor['ram'] < 1024 or flavor['disk'] < 10:
                continue
            flavor_to_use = flavor
            break
        if flavor_to_use:
            LOG.info('Automatically selects flavor %s to instantiate VMs.' %
                     (flavor_to_use['name']))
            self.tenant.kloud.flavor_to_use = flavor_to_use['name']
        else:
            LOG.error('Cannot find a flavor which meets the minimum '
                      'requirements to instantiate VMs.')
            raise KBFlavorCheckException()

        # Nova/Cinder/Neutron quota check
        tenant_id = self.tenant.tenant_id
        meet_quota = True
        for quota_type in ['nova', 'cinder', 'neutron']:
            if quota_type == 'nova':
                quota_manager = base_compute.NovaQuota(self.nova_client, tenant_id)
            elif quota_type == 'cinder':
                quota_manager = base_compute.CinderQuota(self.cinder_client, tenant_id)
            else:
                quota_manager = base_network.NeutronQuota(self.neutron_client, tenant_id)

            meet_quota = True
            quota = quota_manager.get()
            for key, value in self.tenant.tenant_quota[quota_type].iteritems():
                if quota[key] < value:
                    meet_quota = False
                    break

        if not meet_quota:
            LOG.error('%s quota is too small. Minimum requirement: %s.' %
                      (quota_type, self.tenant.tenant_quota[quota_type]))
            raise KBQuotaCheckException()

    def create_resources(self):
        """
        Creates all the User elements associated with a User
        1. Creates the routers
        2. Creates the neutron and nova client objects
        """
        # Create a new neutron client for this User with correct credentials
        creden = {}
        creden['username'] = self.user_name
        creden['password'] = self.password
        creden['auth_url'] = self.tenant.kloud.auth_url
        creden['tenant_name'] = self.tenant.tenant_name

        # Create the neutron client to be used for all operations
        self.neutron_client = neutronclient.Client(**creden)

        # Create a new nova and cinder client for this User with correct credentials
        creden_nova = {}
        creden_nova['username'] = self.user_name
        creden_nova['api_key'] = self.password
        creden_nova['auth_url'] = self.tenant.kloud.auth_url
        creden_nova['project_id'] = self.tenant.tenant_name
        creden_nova['version'] = 2

        self.nova_client = Client(**creden_nova)
        self.cinder_client = cinderclient.Client(**creden_nova)

        if self.tenant.kloud.reusing_tenants:
            self.check_resources_quota()
        else:
            self.update_tenant_quota(self.tenant.tenant_quota)

        config_scale = self.tenant.kloud.scale_cfg

        # Create the user's keypair if configured
        if config_scale.public_key_file:
            self.key_pair = base_compute.KeyPair(self.nova_client)
            self.key_name = self.user_name + '-K'
            self.res_logger.log('keypairs', self.key_name, "")
            self.key_pair.add_public_key(self.key_name, config_scale.public_key_file)

        # Find the external network that routers need to attach to
        external_network = base_network.find_external_network(self.neutron_client)

        # Create the required number of routers and append them to router list
        LOG.info("Creating routers and networks for user %s" % self.user_name)
        for router_count in range(config_scale['routers_per_user']):
            router_instance = base_network.Router(self)
            self.router_list.append(router_instance)
            router_name = self.user_name + "-R" + str(router_count)
            # Create the router and also attach it to external network
            router_instance.create_router(router_name, external_network)
            self.res_logger.log('routers', router_instance.router['router']['name'],
                                router_instance.router['router']['id'])
            # Now create the network resources inside the router
            router_instance.create_network_resources(config_scale)

    def get_first_network(self):
        if self.router_list:
            return self.router_list[0].get_first_network()
        return None

    def get_all_instances(self):
        all_instances = []
        for router in self.router_list:
            all_instances.extend(router.get_all_instances())
        return all_instances



# Copyright 2015 Cisco Systems, Inc.  All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from kb_server import model
from pecan import make_app

def setup_app(config):

    model.init_model()
    app_conf = dict(config.app)

    return make_app(
        app_conf.pop('root'),
        logging=getattr(config, 'logging', {}),
        **app_conf
    )



# Copyright 2015 Cisco Systems, Inc.  All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from collections import deque
from distutils.version import LooseVersion
import threading
import time

import log as logging
import redis

# A set of warned VM version mismatches
vm_version_mismatches = set()

LOG = logging.getLogger(__name__)

class KBVMUpException(Exception):
    pass

class KBSetStaticRouteException(Exception):
    pass

class KBHTTPServerUpException(Exception):
    pass

class KBHTTPBenchException(Exception):
    pass

class KBProxyConnectionException(Exception):
    pass

class KBRunner(object):
    """
    Control the testing VMs on the testing cloud
    """

    def __init__(self, client_list, config, expected_agent_version, single_cloud=True):
        self.client_dict = dict(zip([x.vm_name for x in client_list], client_list))
        self.config = config
        self.single_cloud = single_cloud
        self.result = {}
        self.host_stats = {}
        self.tool_result = {}
        self.expected_agent_version = expected_agent_version
        self.agent_version = None
        self.report = {'seq': 0, 'report': None}

        # Redis
        self.redis_obj = None
        self.pubsub = None
        self.orches_chan_name = "kloudbuster_orches"
        self.report_chan_name = "kloudbuster_report"
        self.message_queue = deque()

    def msg_handler(self):
        for message in self.pubsub.listen():
            if message['data'] == "STOP":
                break
            self.message_queue.append(message)

    def setup_redis(self, redis_server, redis_server_port=6379, timeout=120):
        LOG.info("Setting up the redis connections...")
        connection_pool = redis.ConnectionPool(
            host=redis_server, port=redis_server_port, db=0)

        self.redis_obj = redis.StrictRedis(connection_pool=connection_pool,
                                           socket_connect_timeout=1,
                                           socket_timeout=1)
        success = False
        retry_count = max(timeout / self.config.polling_interval, 1)
        # Check for connections to redis server
        for retry in xrange(retry_count):
            try:
                self.redis_obj.get("test")
                success = True
            except (redis.exceptions.ConnectionError):
                LOG.info("Connecting to redis server... Retry #%d/%d", retry, retry_count)
                time.sleep(self.config.polling_interval)
                continue
            break
        if not success:
            LOG.error("Error: Cannot connect to the Redis server")
            raise KBProxyConnectionException()

        # Subscribe to message channel
        self.pubsub = self.redis_obj.pubsub(ignore_subscribe_messages=True)
        self.pubsub.subscribe(self.report_chan_name)
        self.msg_thread = threading.Thread(target=self.msg_handler)
        self.msg_thread.daemon = True
        self.msg_thread.start()

    def dispose(self):
        self.redis_obj.publish(self.report_chan_name, "STOP")
        self.msg_thread.join()
        if self.pubsub:
            self.pubsub.unsubscribe()
            self.pubsub.close()

    def send_cmd(self, cmd, client_type, data):
        message = {'cmd': cmd, 'sender-id': 'kb-master',
                   'client-type': client_type, 'data': data}
        LOG.kbdebug(message)
        self.redis_obj.publish(self.orches_chan_name, message)

    def polling_vms(self, timeout, polling_interval=None):
        '''
        Polling all VMs for the status of execution
        Guarantee to run once if the timeout is less than polling_interval
        '''
        if not polling_interval:
            polling_interval = self.config.polling_interval
        retry_count = max(timeout / polling_interval, 1)
        retry = cnt_succ = cnt_failed = 0
        clist = self.client_dict.copy()
        samples = []
        http_tool = self.client_dict.values()[0].http_tool

        while (retry < retry_count and len(clist)):
            time.sleep(polling_interval)
            sample_count = 0
            while True:
                try:
                    msg = self.message_queue.popleft()
                except IndexError:
                    # No new message, commands are in executing
                    break

                LOG.kbdebug(msg)
                payload = eval(msg['data'])
                vm_name = payload['sender-id']
                instance = self.client_dict[vm_name]
                cmd = payload['cmd']
                if cmd == 'READY':
                    # If a READY packet is received, the corresponding VM is up
                    # running. We mark the flag for that VM, and skip all READY
                    # messages received afterwards.
                    if instance.up_flag:
                        continue
                    else:
                        clist[vm_name].up_flag = True
                        clist.pop(vm_name)
                        cnt_succ = cnt_succ + 1
                        self.agent_version = payload['data']
                elif cmd == 'REPORT':
                    sample_count = sample_count + 1
                    # Parse the results from HTTP Tools
                    instance = self.client_dict[vm_name]
                    self.result[vm_name] = instance.http_client_parser(**payload['data'])
                    samples.append(self.result[vm_name])
                elif cmd == 'DONE':
                    self.result[vm_name] = payload['data']
                    clist.pop(vm_name)
                    if self.result[vm_name]['status']:
                        # Command returned with non-zero status, command failed
                        LOG.error("[%s] %s", vm_name, self.result[vm_name]['stderr'])
                        cnt_failed = cnt_failed + 1
                    else:
                        # Command returned with zero, command succeed
                        cnt_succ = cnt_succ + 1
                elif cmd == 'DEBUG':
                    LOG.info('[%s] %s' + (vm_name, payload['data']))
                else:
                    LOG.error('[%s] received invalid command: %s' + (vm_name, cmd))

            log_msg = "%d Succeed, %d Failed, %d Pending... Retry #%d" %\
                      (cnt_succ, cnt_failed, len(clist), retry)
            if sample_count != 0:
                log_msg += " (%d sample(s) received)" % sample_count
            LOG.info(log_msg)

            if sample_count != 0:
                report = http_tool.consolidate_samples(samples, len(self.client_dict))
                self.report['seq'] = self.report['seq'] + 1
                self.report['report'] = report
                LOG.info('Periodical report: %s.' % str(self.report))
                samples = []
            retry = retry + 1

        return (cnt_succ, cnt_failed, len(clist))

    def wait_for_vm_up(self, timeout=300):
        cnt_succ = self.polling_vms(timeout)[0]
        if cnt_succ != len(self.client_dict):
            raise KBVMUpException()
        self.send_cmd('ACK', None, None)

    def setup_static_route(self, timeout=30):
        func = {'cmd': 'setup_static_route'}
        self.send_cmd('EXEC', 'http', func)
        cnt_succ = self.polling_vms(timeout)[0]
        if cnt_succ != len(self.client_dict):
            raise KBSetStaticRouteException()

    def check_http_service(self, timeout=30):
        func = {'cmd': 'check_http_service'}
        self.send_cmd('EXEC', 'http', func)
        cnt_succ = self.polling_vms(timeout)[0]
        if cnt_succ != len(self.client_dict):
            raise KBHTTPServerUpException()

    def run_http_test(self):
        func = {'cmd': 'run_http_test'}
        self.send_cmd('EXEC', 'http', func)
        # Give additional 30 seconds for everybody to report results
        timeout = self.config.http_tool_configs.duration + 30
        cnt_pending = self.polling_vms(timeout)[2]
        if cnt_pending != 0:
            LOG.warn("Testing VMs are not returning results within grace period, "
                     "summary shown below may not be accurate!")

        # Parse the results from HTTP Tools
        for key, instance in self.client_dict.items():
            self.result[key] = instance.http_client_parser(**self.result[key])

    def gen_host_stats(self):
        for vm in self.result.keys():
            phy_host = self.client_dict[vm].host
            if phy_host not in self.host_stats:
                self.host_stats[phy_host] = []
            self.host_stats[phy_host].append(self.result[vm])

        http_tool = self.client_dict.values()[0].http_tool
        for phy_host in self.host_stats:
            self.host_stats[phy_host] = http_tool.consolidate_results(self.host_stats[phy_host])

    def run(self):
        try:
            LOG.info("Waiting for agents on VMs to come up...")
            self.wait_for_vm_up()
            if not self.agent_version:
                self.agent_version = "0"
            if (LooseVersion(self.agent_version) != LooseVersion(self.expected_agent_version)):
                # only warn once for each unexpected VM version
                if self.expected_agent_version not in vm_version_mismatches:
                    vm_version_mismatches.add(self.expected_agent_version)
                    LOG.warn("The VM image you are running (%s) is not the expected version (%s) "
                             "this may cause some incompatibilities" %
                             (self.agent_version, self.expected_agent_version))
            if self.single_cloud:
                LOG.info("Setting up static route to reach tested cloud...")
                self.setup_static_route()

            LOG.info("Waiting for HTTP service to come up...")
            self.check_http_service()

            if self.config.prompt_before_run:
                print "Press enter to start running benchmarking tools..."
                raw_input()

            LOG.info("Running HTTP Benchmarking...")
            self.run_http_test()

            # Call the method in corresponding tools to consolidate results
            http_tool = self.client_dict.values()[0].http_tool
            LOG.kbdebug(self.result.values())
            self.tool_result = http_tool.consolidate_results(self.result.values())
            self.tool_result['http_rate_limit'] = self.config.http_tool_configs.rate_limit
            self.tool_result['total_connections'] =\
                len(self.client_dict) * self.config.http_tool_configs.connections
            self.gen_host_stats()
            self.dispose()
        except (KBSetStaticRouteException):
            LOG.error("Could not set static route.")
            self.dispose()
            return
        except (KBHTTPServerUpException):
            LOG.error("HTTP service is not up in testing cloud.")
            self.dispose()
            return
        except KBHTTPBenchException():
            LOG.error("Error in HTTP benchmarking.")
            self.dispose()
            return



# Copyright 2014 Cisco Systems, Inc.  All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

# Module for credentials in Openstack
import getpass
import os
import re

import log as logging

LOG = logging.getLogger(__name__)


class Credentials(object):

    def get_credentials(self):
        dct = {}
        dct['username'] = self.rc_username
        dct['password'] = self.rc_password
        dct['auth_url'] = self.rc_auth_url
        dct['tenant_name'] = self.rc_tenant_name
        return dct

    def get_nova_credentials(self):
        dct = {}
        dct['username'] = self.rc_username
        dct['api_key'] = self.rc_password
        dct['auth_url'] = self.rc_auth_url
        dct['project_id'] = self.rc_tenant_name
        return dct

    def get_nova_credentials_v2(self):
        dct = self.get_nova_credentials()
        dct['version'] = 2
        return dct

    def _init_with_openrc_(self, openrc_contents):
        export_re = re.compile('export OS_([A-Z_]*)="?(.*)')
        for line in openrc_contents.splitlines():
            line = line.strip()
            mstr = export_re.match(line)
            if mstr:
                # get rid of posible trailing double quote
                # the first one was removed by the re
                name = mstr.group(1)
                value = mstr.group(2)
                if value.endswith('"'):
                    value = value[:-1]
                # get rid of password assignment
                # echo "Please enter your OpenStack Password: "
                # read -sr OS_PASSWORD_INPUT
                # export OS_PASSWORD=$OS_PASSWORD_INPUT
                if value.startswith('$'):
                    continue
                # now match against wanted variable names
                if name == 'USERNAME':
                    self.rc_username = value
                elif name == 'AUTH_URL':
                    self.rc_auth_url = value
                elif name == 'TENANT_NAME':
                    self.rc_tenant_name = value

    # Read a openrc file and take care of the password
    # The 2 args are passed from the command line and can be None
    def __init__(self, openrc_file=None, openrc_contents=None, pwd=None, no_env=False):
        self.rc_password = None
        self.rc_username = None
        self.rc_tenant_name = None
        self.rc_auth_url = None
        self.openrc_contents = openrc_contents
        success = True

        if openrc_file:
            if os.path.exists(openrc_file):
                self.openrc_contents = open(openrc_file).read()
            else:
                LOG.error("rc file does not exist %s" % openrc_file)
                success = False
                return

        if self.openrc_contents:
            self._init_with_openrc_(self.openrc_contents)
        elif not no_env:
            # no openrc file passed - we assume the variables have been
            # sourced by the calling shell
            # just check that they are present
            for varname in ['OS_USERNAME', 'OS_AUTH_URL', 'OS_TENANT_NAME']:
                if varname not in os.environ:
                    LOG.warn("%s is missing" % varname)
                    success = False
            if success:
                self.rc_username = os.environ['OS_USERNAME']
                self.rc_auth_url = os.environ['OS_AUTH_URL']
                self.rc_tenant_name = os.environ['OS_TENANT_NAME']

        # always override with CLI argument if provided
        if pwd:
            self.rc_password = pwd
        # if password not know, check from env variable
        elif self.rc_auth_url and not self.rc_password and success:
            if 'OS_PASSWORD' in os.environ and not no_env:
                self.rc_password = os.environ['OS_PASSWORD']
            else:
                # interactively ask for password
                self.rc_password = getpass.getpass(
                    'Please enter your OpenStack Password: ')
        if not self.rc_password:
            self.rc_password = ""
